/*----------------------------------------------------------------------------
;
; phico.c Builds .ico file from one or more source files (.ico, .bmp or .png)
;
; Compiled MSVC 6.0:	cl /Ox phico.c
;
;  Date	 Programmer	Description
; ------ ----------	-----------
; 130507 Paul Houle	Initial version
;
;---------------------------------------------------------------------------*/

#include<stdio.h>
#include<stdlib.h>
#include<string.h>

typedef struct {
	unsigned short const0,const1,imgcnt;
} ICONDIR;

typedef struct {
	unsigned char w,h,colors,const0;
	unsigned short pl, bpp;
	unsigned sz, off;
} ICONDIRENTRY;

typedef struct SFDSTRUC {				// source file image descriptor
	char *fn; unsigned sz; char *pimg;
	struct SFDSTRUC *pnxt;
} SFD;

typedef struct OIDSTRUC {				// output image descriptor
	ICONDIRENTRY ide;
	char *fn; char *pimg;
	struct OIDSTRUC *pnxt;
} OID;

SFD *ReadFile(SFD *psf, char *fn);
OID *ProcessBMP(OID *, SFD *);
OID *ProcessICO(OID *, SFD *);
OID *ProcessPNG(OID *, SFD *);
OID *ProcessImage(OID *, SFD *);
void WriteFile(char *fn, OID *poi);

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; Main function.
*/

int main(int argc, char *argv[]) {
	SFD *psf, sf = {0};					// ptr-to, head-of SFD list
	OID *poi, oi = {0};					// ptr-to, head-of OFD list
	unsigned i;

	if (argc < 3) {
		printf("phico ver 1.0, May 7, 2013 -- paulhoule.com\n"
			"Builds .ico file from input .bmp, .ico or .png\n\n"
			"Usage: phico outfile.ico infile1 (infile2...)\n\n"
			"Examples:\n"
			"  phico nul in.ico - display info about in.ico\n"
			"  phico icon.ico icon.ico new.ico - add new.ico to icon.ico\n"
			"  phico icon.ico in1.bmp in2.png in3.ico - build icon.ico\n"
		);
		exit(-1);
	}

	for (i = 2, psf = &sf; i < argc; i++) {	// load source file(s)
		psf = ReadFile(psf, argv[i]);
	}

	for (psf = &sf, poi = &oi; psf = psf->pnxt; ) {	// process source images
		poi = ProcessImage(poi, psf);
	}

	WriteFile(argv[1], &oi);			// build & output ico file
	return 0;
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; Read file, append its image to SFD list.
*/

SFD *ReadFile(SFD *psf, char *fn) {
	FILE *fh;

	psf = psf->pnxt = calloc(1, sizeof(SFD));	// append new SFD to list
	if (!(fh = fopen(psf->fn = fn, "rb"))		// read file into SFD
	  || fseek(fh, 0, SEEK_END)
	  || ((psf->sz = ftell(fh)) < 16)
	  || !(psf->pimg = malloc(psf->sz))
	  || fseek(fh, 0, SEEK_SET)
	  || fread(psf->pimg, 1, psf->sz, fh) != psf->sz
	  || fclose(fh)) {
		printf("file read error: %s\n", fn); exit(99);
	}
	return psf;
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; Builds and writes output file (.ico format).
*/

void WriteFile(char *fn, OID *poi) {
	FILE *fh;
	char *pbuf, *pimg;
	OID *pwoi = poi;
	ICONDIR id = {0, 1, 0};
	ICONDIRENTRY *pide;
	unsigned totsz = 0;

	while (pwoi = pwoi->pnxt) {			// calc number of images, file size
		id.imgcnt++; totsz += pwoi->ide.sz;

		#define Z256(x) ((x) == 0 ? 256 : (x))
		printf("%2d: %s %3dx%3d%3dbpp%4dclr%7d %s\n", id.imgcnt,
			memcmp(pwoi->pimg, "\x89PNG", 4) ? "bmp" : "png",
			Z256(pwoi->ide.h), Z256(pwoi->ide.w), pwoi->ide.bpp,
			pwoi->ide.colors, pwoi->ide.sz, pwoi->fn);
	}
	totsz += sizeof(id) + id.imgcnt * sizeof(ICONDIRENTRY);
	pbuf = malloc(totsz);

	*(ICONDIR *)pbuf = id;				// create file image
	pide = (ICONDIRENTRY *)(pbuf + sizeof(id));
	pimg = (char *) &pide[id.imgcnt];
	pwoi = poi;
	while (pwoi = pwoi->pnxt) {
		pwoi->ide.off = pimg - pbuf;
		*pide++ = pwoi->ide;
		memcpy(pimg, pwoi->pimg, pwoi->ide.sz);
		pimg += pwoi->ide.sz;
	}

	if (!(fh = fopen(fn, "wb"))			// write output file
	  || fwrite(pbuf, 1, totsz, fh) != totsz || fclose(fh)) {
		printf("file write error: %s\n", fn); exit(99);
	}
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; Processes a single input file image from SFD list.
*/

OID *ProcessImage(OID *poi, SFD *psf) {
	if (!memcmp(psf->pimg, "BM", 2)) {
		poi = ProcessBMP(poi, psf);
	} else if (!memcmp(psf->pimg, "\x89PNG", 4)) {
		poi = ProcessPNG(poi, psf);
	} else if (!memcmp(psf->pimg, "\0\0\1\0", 4)) {
		poi = ProcessICO(poi, psf);
	} else {
		printf("unknown file format: %s\n", psf->fn); exit(99);
	}
	return poi;
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; Append generated ico image to poi list from source bmp image.
; The mask field of the generated ico image will be all zero.
;
; Only a subset of bmp formats are allowed:
;	(1) BITMAPINFOHEADER size of 40 (Windows) or 12 (OS/2)
;	(2) uncompressed (BI_RGB only)
;	(3) planes == 1
;	(4) 0 < width <= 256, 0 < height <= 256 (range of what ico can store)
;
; A few things necessary for ico are handled:
;	(1) Negative height bmp's are flipped
;	(2) Any gap between the BMP header+palette and pixel data is removed
;	(3) 24-bit pixel data is converted to 32-bit
*/

OID *ProcessBMP(OID *poi, SFD *psf) {
	struct { int fsz, res, pxoff; } * const pbfh = (void *) (psf->pimg + 2);
	struct OBI { int ssz; unsigned short w, h, pl, bpp; };
	struct WBI { int ssz, w, h; short pl, bpp; int comp, isz, hr,vr, pcol,ic;
	  } bi = {0};

	if (((struct WBI *)(pbfh + 1))->ssz == sizeof(bi)) {
		memcpy(&bi, pbfh + 1, sizeof(bi));
	} else if (((struct OBI *)(pbfh + 1))->ssz == sizeof(struct OBI)) {
		struct OBI * const pobi = (struct OBI *) (pbfh + 1);
		bi.ssz = sizeof(bi); bi.w = pobi->w;	// translate OS/2 format
		bi.h = pobi->h; bi.pl = pobi->pl; bi.bpp = pobi->bpp;
	}
	if (bi.ssz != sizeof(bi) || bi.comp || bi.pl != 1) {
		printf("unsupported/compressed (or planes not 1): %s\n", psf->fn);
		exit(99);
	}

	if (bi.h < 0) {					// flip bmp in-place if necessary
		int rby = (bi.w * bi.bpp + 31 & ~31) / 8;
		char *pt = malloc(rby);
		char *ps = psf->pimg + pbfh->pxoff;
		char *pd = ps + ((-bi.h) - 1) * rby;
		while (ps < pd) {
			memcpy(pt, ps, rby); memcpy(ps, pd, rby); memcpy(pd, pt, rby);
			ps += rby; pd -= rby;
		}
		free(pt);
		bi.h = -bi.h;
	}

	if ((unsigned)(bi.w-1) > 255 || (unsigned)(bi.h-1) > 255) {
		printf("bad row,col for icon: %d,%d [%s]\n", bi.h, bi.w, psf->fn);
		exit(99);
	}

	if (!bi.pcol && bi.bpp <= 8) bi.pcol = 1 << bi.bpp;

	{
		int obpp = bi.bpp;
		int nbpp = obpp == 24 ? 32 : obpp;
		int orby = (bi.w * obpp + 31 & ~31) / 8;
		int nrby = (bi.w * nbpp + 31 & ~31) / 8;
		char *ps = (char *) (pbfh + 1) + sizeof(bi);
		char *pd;

		poi = poi->pnxt = calloc(1, sizeof(OID)); // append new OID to list
		poi->ide.sz = sizeof(bi) + bi.pcol * 4
			+ bi.h * nrby + bi.h * ((bi.w + 31 & ~31) / 8);
		pd = poi->pimg = calloc(1, poi->ide.sz);

		memcpy(pd, &bi, sizeof(bi));
		((struct WBI *)pd)->bpp = nbpp; ((struct WBI *)pd)->h *= 2;
		pd += sizeof(bi);
		memcpy(pd, ps, bi.pcol * 4);
		pd += bi.pcol * 4;
		ps = psf->pimg + pbfh->pxoff;

		if (nrby == orby) {
			memcpy(pd, ps, bi.h * nrby);
		} else {
			int i = bi.h; do {
				char *pt = ps;
				int j = bi.w; do {
					memcpy(pd, pt, 3); pd += 4; pt += 3;
				} while (--j);
				ps += orby;
			} while (--i);
		}
		poi->ide.w = (char) bi.w; poi->ide.h = (char) bi.h;
		poi->ide.colors = (char) bi.pcol;
		poi->ide.pl = bi.pl; poi->ide.bpp = nbpp;
		poi->fn = psf->fn;
	}
	return poi;
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; Append ico image(s) to poi list from source ico image.
; Not much to do here -- don't have to parse/reformat, format already .ico.
*/

OID *ProcessICO(OID *poi, SFD *psf) {
	ICONDIR * const pid = (void *) psf->pimg;
	ICONDIRENTRY * pide = (void *) (pid + 1);
	int i;

	for (i = 0; i < pid->imgcnt; i++) {
		poi = poi->pnxt = calloc(1, sizeof(OID)); // append new OID to list
		poi->fn = psf->fn;
		poi->pimg = psf->pimg + pide->off;
		poi->ide = *pide++;
	}
	return poi;
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; Append image to poi list from source png image.
*/

unsigned u32(unsigned char *src) {
	return (src[0] << 8 | src[1]) << 16 | (src[2] << 8 | src[3]);
}

OID *ProcessPNG(OID *poi, SFD *psf) {
	char *pimg = psf->pimg;
	unsigned w, h;

	if (memcmp(pimg + 12, "IHDR", 4)) {
		printf("unrecognized PNG file: %s\n", psf->fn);
		exit(99);
	}
	w = u32(pimg + 16); h = u32(pimg + 20);
	if (!w || w > 256 || !h || h > 256) {
		printf("bad row,col for icon: %d,%d [%s]\n", h, w, psf->fn);
		exit(99);
	}

	poi = poi->pnxt = calloc(1, sizeof(OID)); // append new OID to list
	poi->fn = psf->fn;
	poi->pimg = psf->pimg; poi->ide.sz = psf->sz;
	poi->ide.w = (char) w; poi->ide.h = (char) h;
	poi->ide.pl = 1; poi->ide.bpp = pimg[24];

	return poi;
}
